#region Copyright
/*
 * Copyright 2005-2010 the Seasar Foundation and the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
#endregion

using System;
using System.Collections;
using System.Reflection;
using System.Runtime.Remoting.Messaging;
using System.Runtime.Remoting.Proxies;
using Seasar.Framework.Aop.Impl;
using Seasar.Framework.Util;

namespace Seasar.Framework.Aop.Proxy
{
    /// <summary>
    /// AopProxy
    /// </summary>
    /// <remarks>
    /// ߓIvNVɂAOPĂ܂B
    /// </remarks>
    [Serializable]
    public sealed class AopProxy : RealProxy
    {
        /// <summary>
        /// ߓIvNV쐬CX^X
        /// </summary>
        private object _target;

        /// <summary>
        /// KpAspect
        /// </summary>
        private readonly IAspect[] _aspects;

        /// <summary>
        /// ߓIvNV쐬^
        /// </summary>
        private readonly Type _type;

        /// <summary>
        /// \bhƂ̃NX̃CX^XS2ReiɊւ
        /// </summary>
        private readonly Hashtable _parameters;

        public Type TargetType
        {
            get { return _type; }
        }

        /// <summary>
        /// RXgN^
        /// </summary>
        /// <param name="type">Aspect^[Qbǧ^Type</param>
        /// <param name="aspects">KAspect</param>
        /// <param name="parameters">p[^</param>
        /// <param name="target">Aspect^[Qbg</param>
        public AopProxy(Type type, IAspect[] aspects, Hashtable parameters, object target)
            : base(type)
        {
            _type = type;
            _target = target;
            _aspects = aspects;
            _parameters = parameters;
        }

        /// <summary>
        /// RXgN^
        /// </summary>
        /// <param name="type">Aspect^[Qbǧ^Type</param>
        /// <param name="aspects">KAspect</param>
        /// <param name="parameters">p[^</param>
        public AopProxy(Type type, IAspect[] aspects, Hashtable parameters)
            : this(type, aspects, parameters, null)
        {
        }

        /// <summary>
        /// RXgN^
        /// </summary>
        /// <param name="type">Aspect^[Qbǧ^Type</param>
        /// <param name="aspects">KAspect</param>
        public AopProxy(Type type, IAspect[] aspects)
            : this(type, aspects, null)
        {
        }

        /// <summary>
        /// RXgN^
        /// </summary>
        /// <param name="type">Aspect^[Qbǧ^Type</param>
        public AopProxy(Type type)
            : this(type, null)
        {
        }

        /// <summary>
        /// ߓIvNVԂ
        /// </summary>
        /// <returns>ߓIvNṼCX^X</returns>
        public object Create()
        {
            return GetTransparentProxy();
        }

        /// <summary>
        /// ߓIvNVԂ
        /// </summary>
        /// <param name="argTypes">ߓIvNV̑ΏۂƂȂNX̃RXgN^̈̌^̃Xg</param>
        /// <param name="args">ߓIvNV̑ΏۂƂȂNX̃RXgN^̈̃Xg</param>
        /// <returns>ߓIvNṼCX^X</returns>
        public object Create(Type[] argTypes, object[] args)
        {
            ConstructorInfo constructor = ClassUtil.GetConstructorInfo(_type, argTypes);
            _target = ConstructorUtil.NewInstance(constructor, args);
            return GetTransparentProxy();
        }

        /// <summary>
        /// ߓIvNVԂ
        /// </summary>
        /// <param name="argTypes">ߓIvNV̑ΏۂƂȂNX̃RXgN^̈̌^̃Xg</param>
        /// <param name="args">ߓIvNV̑ΏۂƂȂNX̃RXgN^̈̃Xg</param>
        /// <param name="targetType">ߓIvNV̑ΏۂƂȂNX̌^</param>
        /// <returns>ߓIvNṼCX^X</returns>
        public object Create(Type[] argTypes, object[] args, Type targetType)
        {
            ConstructorInfo constructor = ClassUtil.GetConstructorInfo(targetType, argTypes);
            _target = ConstructorUtil.NewInstance(constructor, args);
            return GetTransparentProxy();
        }

        #region RealProxy o

        /// <summary>
        /// AopProxyʂIuWFNg̃\bhsƂ̃\bhĂ΂܂
        /// </summary>
        /// <param name="msg">IMessage</param>
        /// <returns>IMessage</returns>
        /// <seealso="System.Runtime.Remoting.Proxies">System.Runtime.Remoting.Proxies</seealso>
        public override IMessage Invoke(IMessage msg)
        {
            if (_target == null)
            {
                if (!_type.IsInterface) _target = Activator.CreateInstance(_type);
                if (_target == null) _target = new object();
            }

            IMethodMessage methodMessage = msg as IMethodMessage;
            MethodBase method = methodMessage.MethodBase;

            ArrayList interceptorList = new ArrayList();

            if (_aspects != null)
            {
                // `ꂽAspectInterceptor̃Xg̍쐬
                foreach (IAspect aspect in _aspects)
                {
                    IPointcut pointcut = aspect.Pointcut;
                    // IPointcutAdvice(Interceptor)}邩mF
                    if (pointcut == null || pointcut.IsApplied(method))
                    {
                        // AspectKpꍇ
                        interceptorList.Add(aspect.MethodInterceptor);
                    }
                }
            }

            object ret;

            object[] methodArgs;

            if (interceptorList.Count == 0)
            {
                methodArgs = methodMessage.Args;

                try
                {
                    //Interceptor}Ȃꍇ
                    ret = method.Invoke(_target, methodArgs);
                }
                catch (TargetInvocationException ex)
                {
                    // InnerExceptionStackTraceۑ
                    ExceptionUtil.SaveStackTraceToRemoteStackTraceString(ex.InnerException);

                    // InnerExceptionthrow
                    throw ex.InnerException;
                }
            }
            else
            {
                // Interceptor}ꍇ
                IMethodInterceptor[] interceptors = (IMethodInterceptor[])
                    interceptorList.ToArray(typeof(IMethodInterceptor));

                IMethodInvocation invocation = new MethodInvocationImpl(_target,
                    method, methodMessage.Args, interceptors, _parameters);

                ret = interceptors[0].Invoke(invocation);

                methodArgs = invocation.Arguments;
            }

            IMethodReturnMessage mrm = new ReturnMessage(ret, methodArgs, methodArgs.Length,
                methodMessage.LogicalCallContext, (IMethodCallMessage) msg);

            return mrm;
        }

        #endregion

    }
}
